/*=============================================================================
File:	feature_index.cpp
Purpose:	loads the model and calculates the cost for features

Special thanks to Taku Kudo <taku@chasen.org> for orginal implementation in
http://crfpp.sourceforge.net/

Simplifies only for readability and understandability
   Copyright(C) langine team
   email: nlptech360@gmail.com
   The license is under GNU GPL.

=============================================================================*/

#include "feature_index.h"

namespace CRFPP {

static inline char *read_ptr(char **ptr, size_t size)
{
  char *r = *ptr;
  *ptr += size;
  return r;
}

template <class T> static inline void read_static(char **ptr,
                                                  T *value)
{
  char *r = read_ptr(ptr, sizeof(T));
  memcpy(value, r, sizeof(T));
}

int DecoderFeatureIndex::getID(const char *key)
{
	return da_.exactMatchSearch(key);
}

bool DecoderFeatureIndex::open(const char *filename1)
{
    if(!mmap_.open(filename1))
    {
        printf("Cannot open the model\n");
        return false;
    }

  char *ptr = mmap_.begin();
  unsigned int version_ = 0;
  read_static<unsigned int>(&ptr, &version_);

  int type = 0;
  read_static<int>(&ptr, &type);
  read_static<double>(&ptr, &cost_factor_);
  read_static<unsigned int>(&ptr, &maxid_);
  read_static<unsigned int>(&ptr, &xsize_);

  unsigned int dsize = 0;
  read_static<unsigned int>(&ptr, &dsize);

  unsigned int y_str_size;
  read_static<unsigned int>(&ptr, &y_str_size);
  char *y_str = read_ptr(&ptr, y_str_size);
  size_t pos = 0;
  while (pos < y_str_size) {
    y_.push_back(y_str + pos);
    while (y_str[pos++] != '\0') {}
  }

  unsigned int tmpl_str_size;
  read_static<unsigned int>(&ptr, &tmpl_str_size);
  char *tmpl_str = read_ptr(&ptr, tmpl_str_size);
  pos = 0;
  while (pos < tmpl_str_size) {
    char *v = tmpl_str + pos;
    if (v[0] == '\0') {
      ++pos;
    } else if (v[0] == 'U') {
      unigram_templs_.push_back(v);
    } else if (v[0] == 'B') {
      bigram_templs_.push_back(v);
    }
    else 
    {
        printf("unknown type\n");
        return false;
    }

    while (tmpl_str[pos++] != '\0') {}
  }

  da_.set_array(ptr);
  ptr += dsize;

  alpha_float_ = reinterpret_cast<float *>(ptr);
  ptr += sizeof(alpha_float_[0]) * maxid_;

  if(ptr != mmap_.end())
  {
      printf("model file is broken\n");
      return false;
  }

  return true;
}

void DecoderFeatureIndex::clear()
{
  char_freelist_.free();
  feature_cache_.clear();
  for (size_t i = 0; i < thread_num_; ++i)
  {
    node_freelist_[i].free();
    path_freelist_[i].free();
  }
}

char *FeatureIndex::strdup(const char *p)
{
  size_t len = strlen(p);
  char *q = char_freelist_.alloc(len+1);
  strcpy(q, p);
  return q;
}

void FeatureIndex::calcCost(Node *n) {
  n->cost = 0.0;

#define ADD_COST(T, A)                                                  \
  { T c = 0;                                                            \
    for (int *f = n->fvector; *f != -1; ++f) c += (A)[*f + n->y];       \
    n->cost =cost_factor_ *(T)c; }

  if (alpha_float_) ADD_COST(float,  alpha_float_)
    else             ADD_COST(double, alpha_);
#undef ADD_COST
}

void FeatureIndex::calcCost(Path *p) {
  p->cost = 0.0;

#define ADD_COST(T, A)                                          \
  { T c = 0.0;                                                  \
    for (int *f = p->fvector; *f != -1; ++f)                    \
      c += (A)[*f + p->lnode->y * y_.size() + p->rnode->y];     \
    p->cost =cost_factor_*(T)c; }

  if (alpha_float_) ADD_COST(float,  alpha_float_)
    else             ADD_COST(double, alpha_);
}
#undef ADD_COST
}
