/*=============================================================================
File:	tagger.cpp
Purpose:	decodes the sequence with the free model

Special thanks to Taku Kudo <taku@chasen.org> for orginal implementation in
http://crfpp.sourceforge.net/

Simplifies only for readability and understandability
   Copyright(C) langine team
   email: nlptech360@gmail.com
   The license is under GNU GPL.
=============================================================================*/

#include "tagger.h"

namespace CRFPP {

	size_t tokenize2(char *str, const char *del, char* out[], size_t max)
	{
		char *stre = str + strlen(str);
		const char *dele = del + strlen(del);
		size_t size = 0;
		while (size < max)
		{
			char *n = std::find_first_of(str, stre, del, dele);
			*n = '\0';
			if (*str != '\0')
			{
				*out++ = str;
				++size;
			}
			
			if (n == stre) break;
			str = n + 1;
		}
		
		return size;
	}

    void TaggerImpl::close() 
    {
        if (mode_ == TEST) {
            delete feature_index_;
            feature_index_ = 0;
        }
    }

    bool TaggerImpl::add2(size_t size, char *column[])
    {
        size_t xsize = feature_index_->xsize();
        if(mode_ == TEST  && size < xsize)
        {
            printf("# x is small: size= %d\n",xsize);
            return false;
        }

        size_t s = x_.size() + 1;
        x_.resize(s);
        node_.resize(s);
        answer_.resize(s);
        result_.resize(s);
        s = x_.size() - 1;

        for (size_t k = 0; k < size; ++k)
        {
            x_[s].push_back(column[k]);
        }

        result_[s] = answer_[s] = 0;
        node_[s].resize(ysize_);

        return true;
    }

    bool TaggerImpl::clear()
    {
        if (mode_ == TEST) feature_index_->clear();
        x_.clear();
        node_.clear();
        answer_.clear();
        result_.clear();
        Z_ = cost_ = 0.0;

        return true;
    }

    void TaggerImpl::buildLattice()
    {
        if (x_.empty()) return;

        feature_index_->rebuildFeatures(this);

        for (size_t i = 0; i < x_.size(); ++i)
        {
            for (size_t j = 0; j < ysize_; ++j)
            {
                feature_index_->calcCost(node_[i][j]);
                const std::vector<Path *> &lpath = node_[i][j]->lpath;
                for (const_Path_iterator it = lpath.begin(); it != lpath.end(); ++it)
                {
                    feature_index_->calcCost(*it);
                }
            }
        }
    }

    bool TaggerImpl::next()
    {
        while (!agenda_->empty()) {
            QueueElement *top = agenda_->top();
            agenda_->pop();
            Node *rnode = top->node;

            if (rnode->x == 0) {
                for (QueueElement *n = top; n; n = n->next)
                    result_[n->node->x] = n->node->y;
                cost_ = top->gx;
                return true;
            }

            for (const_Path_iterator it = rnode->lpath.begin();
                it != rnode->lpath.end(); ++it) {
                    QueueElement *n =nbest_freelist_->alloc();
                    n->node = (*it)->lnode;
                    n->gx   = -(*it)->lnode->cost     -(*it)->cost +  top->gx;
                    n->fx   = -(*it)->lnode->bestCost -(*it)->cost +  top->gx;
                    //          |              h(x)                 |  |  g(x)  |
                    n->next = top;
                    agenda_->push(n);
            }
        }

        return 0;
    }

    void TaggerImpl::viterbi() {
        for (size_t i = 0;   i < x_.size(); ++i) {
            for (size_t j = 0; j < ysize_; ++j) {
                double bestc = -1e37;
                Node *best = 0;
                const std::vector<Path *> &lpath = node_[i][j]->lpath;
                for (const_Path_iterator it = lpath.begin(); it != lpath.end(); ++it) {
                    double cost = (*it)->lnode->bestCost +(*it)->cost +
                        node_[i][j]->cost;
                    if (cost > bestc) {
                        bestc = cost;
                        best  = (*it)->lnode;
                    }
                }
                node_[i][j]->prev     = best;
                node_[i][j]->bestCost = best ? bestc : node_[i][j]->cost;
            }
        }

        double bestc = -1e37;
        Node *best = 0;
        size_t s = x_.size()-1;
        for (size_t j = 0; j < ysize_; ++j) {
            if (bestc < node_[s][j]->bestCost) {
                best  = node_[s][j];
                bestc = node_[s][j]->bestCost;
            }
        }

        for (Node *n = best; n; n = n->prev)
            result_[n->x] = n->y;

        cost_ = -node_[x_.size()-1][result_[x_.size()-1]]->bestCost;
    }

    const char* TaggerImpl::toString() 
    {
        os_.assign("");

#define PRINT                                                   \
        for (size_t i = 0; i < x_.size(); ++i) {                      \
        for (std::vector<const char*>::iterator it = x_[i].begin(); \
        it != x_[i].end(); ++it)                               \
        os_ << *it << '\t';                                       \
        os_ << yname(y(i));                                         \
        if (vlevel_ >= 1) os_ << '/' << prob(i);                    \
        if (vlevel_ >= 2) {                                         \
        for (size_t j = 0; j < ysize_; ++j)                       \
        os_ << '\t' << yname(j) << '/' << prob(i, j);           \
        }                                                           \
        os_ << '\n';                                                \
        }                                                             \

        os_ << '\n';
        if (nbest_ >= 1)
        {
            for (size_t n = 0; n < nbest_; ++n)
            {
                if (!next()) break;
                os_ << "# " << n << " " << prob() << '\n';
                PRINT;
            }
        }
        else
        {
            if (vlevel_ >= 1) os_ << "# " << prob() << '\n';
            PRINT;
        }

        return const_cast<const char*>(os_.c_str());
#undef PRINT
    }

    bool TaggerImpl::AddLine(char* pszLine)
    {
        char* column[256];
        char *pszColyLine = feature_index_->strdup(pszLine);
        size_t size = tokenize2(pszColyLine, "\t ", column, sizeof(column));
        if (!add2(size, column)) return false;

        return true;
    }

    bool TaggerImpl::ReadSequence(FILE* fpTest)
    {
        char szLine[256];
        clear();

        while(!feof(fpTest))
        {
            memset(szLine,0,256*sizeof(char));
            fgets(szLine,256,fpTest);

            char* pch = strrchr(szLine,'\r');
            if(pch != NULL) *pch = 0;
            pch = strrchr(szLine,'\n');
            if(pch != NULL) *pch = 0;

            if (szLine[0] == '\0' || szLine[0] == ' ' || szLine[0] == '\t') break;

            if (!AddLine(szLine)) return false;
        }

        return true;
    }

    bool TaggerImpl::ParseSequence()
    {
        if(!feature_index_->buildFeatures(this))
            return false;

        if (x_.empty()) return true;

        buildLattice();

        viterbi();

        return true;
    }

    bool TaggerImpl::ProcessingFile(FILE* fpInput, FILE* fpOutput)
    {
        if(!ReadSequence(fpInput))
            return false;

        if(!ParseSequence())
            return false;

        if(x_.empty()) return true;

        toString();

        const char* pszResult = os_.data();
        if(pszResult != NULL)
        {
            fprintf(fpOutput,"%s\n",pszResult);
        }

        return true;
    }

    bool TaggerImpl::open(const char* pszModelPath)
    {
        feature_index_ = new DecoderFeatureIndex();

        if(!feature_index_->open(pszModelPath))
            return false;

        ysize_ = feature_index_->ysize();

        return true;
    }
}
